@@ -1,4 +1,4 @@
-/*	$OpenBSD: bgpctl.c,v 1.283 2022/08/31 15:00:53 claudio Exp $ */
+/*	$OpenBSD: bgpctl.c,v 1.290 2023/03/13 16:59:22 claudio Exp $ */
 
 /*
  * Copyright (c) 2003 Henning Brauer <henning@openbsd.org>
@@ -79,7 +79,7 @@ int
 main(int argc, char *argv[])
 {
 	struct sockaddr_un	 sa_un;
-	int			 fd, n, done, ch, verbose = 0;
+	int			 fd, n, done, numdone, ch, verbose = 0;
 	struct imsg		 imsg;
 	struct network_config	 net;
 	struct parse_result	*res;
@@ -256,6 +256,12 @@ main(int argc, char *argv[])
 	case SHOW_RIB_MEM:
 		imsg_compose(ibuf, IMSG_CTL_SHOW_RIB_MEM, 0, 0, -1, NULL, 0);
 		break;
+	case SHOW_METRICS:
+		output = &ometric_output;
+		numdone = 2;
+		imsg_compose(ibuf, IMSG_CTL_SHOW_NEIGHBOR, 0, 0, -1, NULL, 0);
+		imsg_compose(ibuf, IMSG_CTL_SHOW_RIB_MEM, 0, 0, -1, NULL, 0);
+		break;
 	case RELOAD:
 		imsg_compose(ibuf, IMSG_CTL_RELOAD, 0, 0, -1,
 		    res->reason, sizeof(res->reason));
@@ -366,18 +372,14 @@ main(int argc, char *argv[])
 		break;
 	}
 
-	while (ibuf->w.queued)
-		if (msgbuf_write(&ibuf->w) <= 0 && errno != EAGAIN)
-			err(1, "write error");
-
 	output->head(res);
 
-	while (!done) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			err(1, "imsg_read error");
-		if (n == 0)
-			errx(1, "pipe closed");
+ again:
+	while (ibuf->w.queued)
+		if (msgbuf_write(&ibuf->w) <= 0)
+			err(1, "write error");
 
+	while (!done) {
 		while (!done) {
 			if ((n = imsg_get(ibuf, &imsg)) == -1)
 				err(1, "imsg_get error");
@@ -387,6 +389,20 @@ main(int argc, char *argv[])
 			done = show(&imsg, res);
 			imsg_free(&imsg);
 		}
+
+		if (done)
+			break;
+
+		if ((n = imsg_read(ibuf)) == -1)
+			err(1, "imsg_read error");
+		if (n == 0)
+			errx(1, "pipe closed");
+
+	}
+
+	if (res->action == SHOW_METRICS && --numdone > 0) {
+		done = 0;
+		goto again;
 	}
 
 	output->tail();
@@ -416,21 +432,29 @@ show(struct imsg *imsg, struct parse_result *res)
 
 	switch (imsg->hdr.type) {
 	case IMSG_CTL_SHOW_NEIGHBOR:
+		if (output->neighbor == NULL)
+			break;
 		p = imsg->data;
 		output->neighbor(p, res);
 		break;
 	case IMSG_CTL_SHOW_TIMER:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(t))
 			errx(1, "wrong imsg len");
+		if (output->timer == NULL)
+			break;
 		memcpy(&t, imsg->data, sizeof(t));
 		if (t.type > 0 && t.type < Timer_Max)
 			output->timer(&t);
 		break;
 	case IMSG_CTL_SHOW_INTERFACE:
+		if (output->interface == NULL)
+			break;
 		iface = imsg->data;
 		output->interface(iface);
 		break;
 	case IMSG_CTL_SHOW_NEXTHOP:
+		if (output->nexthop == NULL)
+			break;
 		nh = imsg->data;
 		output->nexthop(nh);
 		break;
@@ -438,18 +462,24 @@ show(struct imsg *imsg, struct parse_result *res)
 	case IMSG_CTL_SHOW_NETWORK:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(*kf))
 			errx(1, "wrong imsg len");
+		if (output->fib == NULL)
+			break;
 		kf = imsg->data;
 		output->fib(kf);
 		break;
 	case IMSG_CTL_SHOW_FIB_TABLES:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(*kt))
 			errx(1, "wrong imsg len");
+		if (output->fib_table == NULL)
+			break;
 		kt = imsg->data;
 		output->fib_table(kt);
 		break;
 	case IMSG_CTL_SHOW_RIB:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(rib))
 			errx(1, "wrong imsg len");
+		if (output->rib == NULL)
+			break;
 		memcpy(&rib, imsg->data, sizeof(rib));
 		aslen = imsg->hdr.len - IMSG_HEADER_SIZE - sizeof(rib);
 		asdata = imsg->data;
@@ -462,6 +492,8 @@ show(struct imsg *imsg, struct parse_result *res)
 			warnx("bad IMSG_CTL_SHOW_RIB_COMMUNITIES received");
 			break;
 		}
+		if (output->communities == NULL)
+			break;
 		output->communities(imsg->data, ilen, res);
 		break;
 	case IMSG_CTL_SHOW_RIB_ATTR:
@@ -470,23 +502,31 @@ show(struct imsg *imsg, struct parse_result *res)
 			warnx("bad IMSG_CTL_SHOW_RIB_ATTR received");
 			break;
 		}
+		if (output->attr == NULL)
+			break;
 		output->attr(imsg->data, ilen, res->flags, 0);
 		break;
 	case IMSG_CTL_SHOW_RIB_MEM:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(stats))
 			errx(1, "wrong imsg len");
+		if (output->rib_mem == NULL)
+			break;
 		memcpy(&stats, imsg->data, sizeof(stats));
 		output->rib_mem(&stats);
 		return (1);
 	case IMSG_CTL_SHOW_SET:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(set))
 			errx(1, "wrong imsg len");
+		if (output->set == NULL)
+			break;
 		memcpy(&set, imsg->data, sizeof(set));
 		output->set(&set);
 		break;
 	case IMSG_CTL_SHOW_RTR:
 		if (imsg->hdr.len < IMSG_HEADER_SIZE + sizeof(rtr))
 			errx(1, "wrong imsg len");
+		if (output->rtr == NULL)
+			break;
 		memcpy(&rtr, imsg->data, sizeof(rtr));
 		output->rtr(&rtr);
 		break;
@@ -495,6 +535,8 @@ show(struct imsg *imsg, struct parse_result *res)
 			warnx("got IMSG_CTL_RESULT with wrong len");
 			break;
 		}
+		if (output->result == NULL)
+			break;
 		memcpy(&rescode, imsg->data, sizeof(rescode));
 		output->result(rescode);
 		return (1);
@@ -669,7 +711,7 @@ fmt_flags(uint32_t flags, int sum)
 	if (sum) {
 		if (flags & F_PREF_INVALID)
 			*p++ = 'E';
-		if (flags & F_PREF_OTC_LOOP)
+		if (flags & F_PREF_OTC_LEAK)
 			*p++ = 'L';
 		if (flags & F_PREF_ANNOUNCE)
 			*p++ = 'A';
@@ -695,8 +737,8 @@ fmt_flags(uint32_t flags, int sum)
 
 		if (flags & F_PREF_INVALID)
 			strlcat(buf, ", invalid", sizeof(buf));
-		if (flags & F_PREF_OTC_LOOP)
-			strlcat(buf, ", otc loop", sizeof(buf));
+		if (flags & F_PREF_OTC_LEAK)
+			strlcat(buf, ", otc leak", sizeof(buf));
 		if (flags & F_PREF_STALE)
 			strlcat(buf, ", stale", sizeof(buf));
 		if (flags & F_PREF_ELIGIBLE)
@@ -729,6 +771,19 @@ fmt_ovs(uint8_t validation_state, int sum)
 	}
 }
 
+const char *
+fmt_avs(uint8_t validation_state, int sum)
+{
+	switch (validation_state) {
+	case ASPA_INVALID:
+		return (sum ? "!" : "invalid");
+	case ASPA_VALID:
+		return (sum ? "V" : "valid");
+	default:
+		return (sum ? "?" : "unknown");
+	}
+}
+
 const char *
 fmt_mem(long long num)
 {
@@ -989,19 +1044,19 @@ fmt_ext_community(uint8_t *data)
 		ext = be64toh(ext) & 0xffffffffffffLL;
 		switch (ext) {
 		case EXT_COMMUNITY_OVS_VALID:
-			snprintf(buf, sizeof(buf), "%s valid ",
+			snprintf(buf, sizeof(buf), "%s valid",
 			    log_ext_subtype(type, subtype));
 			return buf;
 		case EXT_COMMUNITY_OVS_NOTFOUND:
-			snprintf(buf, sizeof(buf), "%s not-found ",
+			snprintf(buf, sizeof(buf), "%s not-found",
 			    log_ext_subtype(type, subtype));
 			return buf;
 		case EXT_COMMUNITY_OVS_INVALID:
-			snprintf(buf, sizeof(buf), "%s invalid ",
+			snprintf(buf, sizeof(buf), "%s invalid",
 			    log_ext_subtype(type, subtype));
 			return buf;
 		default:
-			snprintf(buf, sizeof(buf), "%s 0x%llx ",
+			snprintf(buf, sizeof(buf), "%s 0x%llx",
 			    log_ext_subtype(type, subtype),
 			    (unsigned long long)ext);
 			return buf;
@@ -1020,6 +1075,8 @@ const char *
 fmt_set_type(struct ctl_show_set *set)
 {
 	switch (set->type) {
+	case ASPA_SET:
+		return "ASPA";
 	case ROA_SET:
 		return "ROA";
 	case PREFIX_SET:
@@ -1332,7 +1389,7 @@ print_afi(u_char *p, uint8_t len)
 	/* safi, 1 byte */
 	memcpy(&safi, p, sizeof(safi));
 	if (afi2aid(afi, safi, &aid) == -1)
-		printf("unkown afi %u safi %u", afi, safi);
+		printf("unknown afi %u safi %u", afi, safi);
 	else
 		printf("%s", aid2str(aid));
 }
@@ -1456,7 +1513,7 @@ show_mrt_capabilities(u_char *p, uint16_t len)
 		len -= capa_len;
 	}
 	if (len != 0) {
-		printf("length missmatch while capability parsing");
+		printf("length mismatch while capability parsing");
 		return (-1);
 	}
 	return (totlen);
